from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection


import pytest


import os
from unittest.mock import Mock


class TestDebugExecutionCommands:
    old_dir = os.getcwd()

    def setup_method(self):
        os.chdir(os.path.dirname(__file__))

    def teardown_method(self):
        os.chdir(self.old_dir)


    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c

    @pytest.fixture
    def dbg(self,conn):
        dbg = CommandLineDebugger(conn)
        assert dbg.handle("load example")
        return dbg

    
    def test_continue(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400611")
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Address Filter: 0x400611.")

    
    def test_set_start_address(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400611")
        conn.send_info.assert_called_with("Address 0x400611 added to breakpoints.")
        assert dbg.handle("add_breakpoint 0x4005de")
        assert dbg.handle("set_entry_state 0x40054d")
        conn.send_info.assert_called_with("Execution will start at address 0x40054d.")
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Address Filter: 0x4005de.") # should not stop at 0x400611
    # - start
    
    def test_start(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x400611")
        assert dbg.handle("run")
        conn.send_info.assert_called_with("Break: Address Filter: 0x400611.")

    # - step into
    
    def test_step_into(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x40062a")
        assert dbg.handle("run")
        assert dbg.handle("step")
        conn.send_info.assert_called_with("Stepped to: 0x40054d.")

    # - step out
    
    def test_step_out(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x40054d")
        assert dbg.handle("run")
        assert dbg.handle("step_out")
        conn.send_info.assert_called_with("Stepped to: 0x40062f.")

    # - step over
    
    def test_step_over(self, dbg, conn):
        assert dbg.handle("add_breakpoint 0x40062a")
        assert dbg.handle("run")
        assert dbg.handle("step_over")
        conn.send_info.assert_called_with("Stepped to: 0x40062f.")



from unittest.mock import Mock, Mock
import pytest

from dAngr.cli.command_line_debugger import CommandLineDebugger
from dAngr.cli.cli_connection import CliConnection
import subprocess



class TestExamples:


    @pytest.fixture
    def conn(self):
        c = CliConnection()
        c.send_result = Mock()
        c.send_info = Mock()
        c.send_error = Mock()
        return c


    # results = {
    #     '00_angr_find.md': b'YXIACZSW',
    #     '01_angr_avoid.md': b'JLVUSGJZ',
    #     '02_angr_find_condition.md': b'OHYJUMBE',
    #     '03_angr_symbolic_registers.md': 'e9b37483 7aab5fde 8f5b48ea',
    #     '04_angr_symbolic_stack.md':'2089710965 12847883',
    #     '05_angr_symbolic_memory.md': 'OJQVXIVX LLEAOODW UVCWUVVC AJXJMVKA',
    #     '06_angr_symbolic_dynamic_memory.md':'OFIJHOXV FBQISOZO',
    #     "07_angr_symbolic_file.md": 'OBAXRUZT',
    #     "08_angr_constraints.md": 'ZEVKWROAYILRPZYB',
    #     "09_angr_hooks.md": b'QREPXOHPJPOQKQLKNOBMULEMGMLNHNIH',
    #     "10_angr_simprocedures.md": b'MTMDRONBBNSAAMNS',
    #     "11_angr_sim_scanf.md": '1447907916 1146768724',
    #     "13_angr_static_binary.md": b'EADQYLAR',
    #     "14_angr_shared_library.md": 'TSDLQKWZ',
    #     "15_angr_arbitrary_read.md": '2358019 AAAAAAAAAAAAAAAAWISO'
    # }

    # 
    # def test_oregonctf_examples(self, conn):
    #     # for each .md file in the examples/malware.oregonctf.org directory run the script
    #     # and check the output

    #     for file in sorted(os.listdir("examples/malware.oregonctf.org")):
    #         if file.endswith(".md"):
    #             print(file)
    #             dbg = CommandLineDebugger(conn)
    #             dbg.handle("run_script 'examples/malware.oregonctf.org/" + file + "'")
    #             assert conn.send_result.call_args[0][0] == self.results[file], f"Failed on {file}"
    #             conn.send_result.reset_mock()
    #             conn.send_info.reset_mock()
    #             conn.send_error.reset_mock()
    
    
    def test_00_angr_find(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/00_angr_find.md'")
        assert conn.send_result.call_args[0][0] == b'YXIACZSW'
    
    
    def test_01_angr_avoid(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/01_angr_avoid.md'")
        assert conn.send_result.call_args[0][0] == b'JLVUSGJZ'

    
    def test_02_angr_find_condition(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/02_angr_find_condition.md'")
        assert conn.send_result.call_args[0][0] == b'OHYJUMBE'
    
    
    def test_03_angr_symbolic_registers(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/03_angr_symbolic_registers.md'")
        assert conn.send_result.call_args[0][0] == 'e9b37483 7aab5fde 8f5b48ea'

    
    def test_04_angr_symbolic_stack(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/04_angr_symbolic_stack.md'")
        assert conn.send_result.call_args[0][0] == '2089710965 12847883'
    
    
    def test_05_angr_symbolic_memory(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/05_angr_symbolic_memory.md'")
        assert conn.send_result.call_args[0][0] == 'OJQVXIVX LLEAOODW UVCWUVVC AJXJMVKA'

    
    def test_06_angr_symbolic_dynamic_memory(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/06_angr_symbolic_dynamic_memory.md'")
        assert conn.send_result.call_args[0][0] == 'OFIJHOXV FBQISOZO'

    
    def test_07_angr_symbolic_file(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/07_angr_symbolic_file.md'")
        assert conn.send_result.call_args[0][0] == 'OBAXRUZT'

    
    def test_08_angr_constraints(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/08_angr_constraints.md'")
        assert conn.send_result.call_args[0][0] == 'ZEVKWROAYILRPZYB'

    
    def test_09_angr_hooks(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/09_angr_hooks.md'")
        assert conn.send_result.call_args[0][0] == b'QREPXOHPJPOQKQLKNOBMULEMGMLNHNIH'

    
    def test_10_angr_simprocedures(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/10_angr_simprocedures.md'")
        assert conn.send_result.call_args[0][0] == b'MTMDRONBBNSAAMNS'

    
    def test_11_angr_sim_scanf(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/11_angr_sim_scanf.md'")
        assert conn.send_result.call_args[0][0] == '1447907916 1146768724'

    
    def test_13_angr_static_binary(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/13_angr_static_binary.md'")
        assert conn.send_result.call_args[0][0] == b'EADQYLAR'

    
    def test_14_angr_shared_library(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/14_angr_shared_library.md'")
        assert conn.send_result.call_args[0][0] == 'TSDLQKWZ'

    def test_15_angr_arbitrary_read(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/15_angr_arbitrary_read.md'")
        assert conn.send_result.call_args[0][0][:7] == '2358019'
        assert conn.send_result.call_args[0][0][-5:-1] == 'WISO'

    
    def test_16_angr_arbitrary_write(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/16_angr_arbitrary_write.md'")
        assert conn.send_result.call_args[0][0][:7] == '6712341'
        assert conn.send_result.call_args[0][0][8:16] == 'NEDVTNOP'
        assert conn.send_result.call_args[0][0][-4:] == '<RCM'

    
    def test_17_angr_arbitrary_jump(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/malware.oregonctf.org/17_angr_arbitrary_jump.md'")
        assert conn.send_result.call_args[0][0] == 'CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCPRCMCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC'
    
    def test_crackme(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/01_easy/crackme.md'")
        assert conn.send_result.call_args[0][0] == "b'ais3{I_tak3_g00d_n0t3s}'"
    

    def test_defcamp_re100(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/01_easy/defcamp_re100.md'")
        assert conn.send_result.call_args[0][0] == "b'Code_Talkers'"
    
    def test_fauxware(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/01_easy/fauxware.md'")
        assert conn.send_result.call_args[0][0] == "b'SOSNEAKY'"

    def test_flareon2015_2(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/01_easy/flareon2015_2.md'")
        assert conn.send_result.call_args[0][0] == "b'a_Little_b1t_harder_plez@flare-on.com'"

    def test_mma_howtouse(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/01_easy/mma_howtouse.md'")
        assert conn.send_result.call_args[0][0] == "MMA{fc7d90ca001fc8712497d88d9ee7efa9e9b32ed8}"
    
    def test_securityfest_fairlight(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/01_easy/securityfest_fairlight.md'")
        assert conn.send_result.call_args[0][0] == b'4ngrman4gem3nt'
    
    def test_asisctffinal2015_fake(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/02_medium/asisctffinals2015_fake.md'")
        assert conn.send_result.call_args[0][0] == "b'ASIS{f5f7af556bd6973bd6f2687280a243d9}'"
    
    def test_defcon2016quals_baby_re(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/02_medium/defcon2016quals_baby-re.md'")
        assert conn.send_result.call_args[0][0] == 'Math is hard!'

    def test_flareon2015_5(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/02_medium/flareon2015_5.md'")
        assert conn.send_result.call_args[0][0] == "b'Sp1cy_7_layer_OSI_dip@flare-on.com'"

    def test_google2016_unbreakable_1(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/02_medium/google2016_unbreakable.md'")
        assert conn.send_result.call_args[0][0] == 'CTF{0The1Quick2Brown3Fox4Jumped5Over6The7Lazy8Fox9}'

    def test_hackcon2016_angry_reverser(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/02_medium/hackcon2016_angry-reverser.md'")
        assert conn.send_result.call_args[0][0] == b'HACKCON{VVhYS04ngrY}'

    def test_0ctf_trace(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/03_hard/0cft_trace.md'")
        assert conn.send_result.call_args[0][0] == "b'0ctf{I_l0v3_4ngr!}'"
    
    def test_whitehat_crypto400(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/Reverse/03_hard/whitehat_crypto400.md'")
        assert conn.send_result.call_args[0][0] == "b'WhiteHat{I_Love_Angr_And_Z3_Solver}'"
    
    def test_cgc_identification(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/vuln_exploit/01_easy/cgc_identification.md'")
        exploit_str = eval(conn.send_result.call_args[0][0])
        assert len(exploit_str) >= 92
        assert chr(exploit_str[0]) == '^'
    
    def test_insomnihack_aeg(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/vuln_exploit/02_medium/insomnihack_aeg.md'")
        exploit_str = eval(conn.send_result.call_args[0][0])
        # (cat file; echo echo BUMU) | ./examples/angr_examples/repo/angr_examples/examples/insomnihack_aeg/demo_bi
        print_exploit_str = f"python -c 'import sys; sys.stdout.buffer.write({str(exploit_str).replace('\'', '"')})'"
        test = subprocess.check_output(f'({print_exploit_str}; echo echo pwned) | ./examples/angr_examples/vuln_exploit/02_medium/demo_bin', shell=True)
        assert test == b'pwned\n'
    
    def test_mbrainfuzz(self, conn):
        dbg = CommandLineDebugger(conn)
        dbg.handle("run_script 'examples/angr_examples/vuln_exploit/03_hard/mbrainfuzz.md'")
        





from typing import cast
from dAngr.cli.cli_connection import CliConnection
from dAngr.cli.debugger_commands.base import BaseCommand
from dAngr.cli.grammar.expressions import ReferenceObject
from dAngr.cli.less_view import Less


class ViewerCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)


    def less(self):
        """
        View previously retrieved output in a scrollable pane.

        Short name: less
        """
        out = cast(CliConnection,self.debugger.conn).output
        if not out and len(cast(CliConnection,self.debugger.conn).history) > 1:
            out = cast(CliConnection,self.debugger.conn).history[-2]

        Less().show_less(out)

    def print(self, text:object|ReferenceObject):
        """
        Print a text.

        Args:
            text (object|ReferenceObject): Text or object content to print.

        Short name: pr
        """
        self.send_result(str(text), False)
        
    def println(self, text:object|ReferenceObject = ""):
        """
        Print a text, ending with a newline

        Args:
            text (object|ReferenceObject): Text or object content to print.

        Short name: prnl
        """
        self.send_result(str(text), True)
    
    def history(self, n:int=10):
        """
        Print the last n commands.

        Args:
            n (int): Number of commands to print. Default is 10.

        Short name: h
        """
        # print last n commands
        return cast(CliConnection,self.debugger.conn).history
        for i in range(1, n+1):
            try:
                index = len(history) - i
                hs = history[-i]
                if not hs:
                    continue
                h = "\n".join([str(s) for s in hs[0] if s])
                #send first 70 chars of h
                self.send_result(f"${index}: {h[:70]}")
            except IndexError:
                break
    def clear_history(self):
        """
        Clear the command history.

        Short name: ch
        """
        cast(CliConnection,self.debugger.conn).clear_history()
        self.send_info("Command history cleared.")


import angr
import claripy

from dAngr.angr_ext.models import BasicBlock
from dAngr.cli.models import Register


class PrettyPrint():
    def print(self, obj, indent):
        return " "* indent + str(obj)


class StatePrinter(PrettyPrint):
    def print(self, obj, indent):
        state:angr.SimState = obj
        return " "*indent + f"State: {hex(state.addr)}"
    
    
class ListPrinter(PrettyPrint):
    def print(self, obj, indent):
        if len(obj) == 0:
            return "[]"
        #if the elements are objects, convert them to strings with a newline in between
        if all([isinstance(e, object) for e in obj]):
            return "\n".join([pretty_print(e, indent) for e in obj])
        return f"[{', '.join([pretty_print(e, indent) for e in obj])}]"
class DictionaryPrinter(PrettyPrint):
    def print(self, obj, indent):
        if len(obj) == 0:
            return "{}"
        if all([isinstance(e, object) for e in obj]):
            #if the values are objects, print each entry as a name colon string rep of the value with a newline in between. End with a curly brace, each entry on a new line
            return "{" + "\n".join([f"{k}: {pretty_print(v, indent)}" for k,v in obj.items()]) + "}"
        #if the values are bae types, print the dictionary with curly braces
        return "{" + ", ".join([f"{k}: {v}" for k,v in obj.items()]) + "}"
class BasicBlockPrinter(PrettyPrint):
    def print(self, obj, indent):
        bb:BasicBlock = obj
        #print address, size, number of instructions and disassembly
        indentation = " "*indent
        return f"{indentation}Address: {hex(bb.address)}:\n{indentation}Size: {bb.size} bytes\n{indentation}Number of Instructions: {bb.instructions}\n{indentation}Disassembly:\n{pretty_print(bb.assembly, indent+2)}"

class RegisterPrinter(PrettyPrint):
    def print(self, obj, indent):
        reg:Register = obj
        v = reg.value if not reg.value.concrete else hex(reg.value.concrete_value)
        return f"{reg.name} (offset: {hex(reg.offset)}, size: {reg.size} bytes): {pretty_print(v)})"

class BVPrettyPrint(PrettyPrint):
    def print(self, obj, indent):
        bv:claripy.ast.BV = obj
        if bv.concrete:
            return f"{hex(bv.concrete_value)}"
        else:
            return f"BV with size:{bv.size()}"
list_printers = {
    angr.SimState:StatePrinter(), 
                 list:ListPrinter(), 
                 dict:DictionaryPrinter(), 
                 BasicBlock:BasicBlockPrinter(),
                Register:RegisterPrinter(),
                claripy.ast.BV:BVPrettyPrint()
                 }

def pretty_print(obj, indent=0):
    if obj is None:
        return "None"
    #if data is a list:
    #   if th elements are objects, convert them to strings with a newline in between
    #   if the elements are base types (integers, strings, bools, ...) convert them to a string, print the list with square brackets
    #if data is a dictionary:
    #   if the values are bae types, print the dictionary with curly braces
    #   if the values are objects, print each entry as a name colon string rep of the value with a newline in between. End with a curly brace, each entry on a new line
    #if data is an object, print the object, unless a template is provided
    return list_printers.get(type(obj), PrettyPrint()).print(obj, indent+2)    



class CommandError(Exception):
    """Base for command execution errors."""
    pass
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, CommandError):
            return False
        return self.args == value.args


import os

from dAngr.cli.models import State
from .base import BaseCommand
from dAngr.exceptions.DebuggerCommandError import DebuggerCommandError
from prompt_toolkit.shortcuts import ProgressBar
from dAngr.utils.loggers import dAngr_log_config

class GeneralCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)

    def help(self, command: str|None = None):
        """
        Show dAngr help.

        Usage: help [command]

        Args:
        command (str|None): The command to show help for. If not provided, show general help.

        Short name: ?

        """
        if command:
            spec = self.debugger.get_command_spec(command)
            if spec is None:
                self.send_error(f"Command '{command}' not found. Type 'help' or '?' for a list of available commands.")
            else:
                self.debugger.list_args(spec)
        else:
             self.debugger.list_commands()

    def set_log_level(self, module:str, level:str = "DEBUG"):
        """
        Set the log level for a module.

        Usage: set_log_level module level

        Args:
        module (str): The module to set the log level for.
        level (str): The log level to set. Default is DEBUG. Valid values are "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET".

        Short name: log

        """
        dAngr_log_config.set_module(module, level)
        self.send_info(f"Set log level for module '{module}' to {level}")
        
    def list_loggers(self):
        """
        List all loggers.

        Usage: list_loggers

        Short name: ll

        """
        return dAngr_log_config.list_modules_from_loggers()



from dAngr.angr_ext.debugger import Debugger
from dAngr.cli.debugger_commands import BaseCommand
from dAngr.cli.filters import AddressFilter, AndFilterList, Filter, FilterFunction, FilterList, FunctionFilter, OrFilterList, SourceFilter, StdStreamFilter
from dAngr.exceptions import DebuggerCommandError, ExecutionError
from dAngr.utils.utils import StreamType

class FilterCommands(BaseCommand):

    def __init__(self, debugger:Debugger):
        super().__init__(debugger)
        

    def filter(self, exclude:bool = False, *filters:Filter):
        """
        Add filters to the list of breakpoints or exclusions.

        Args:
            filters (tuple): The filters to add. Can be an address, source file and line number, function name or stream text.
            exclude (bool): True to add to exclusions, False to add to breakpoints.
        
        Short name: fae
        """
        if len(filters) == 0:
            raise DebuggerCommandError("At least one filter must be provided.")
        lst = self.debugger.exclusions if exclude else self.debugger.breakpoints
        if len(filters) == 1:
            lst.append(filters[0])
        else:
            lst.append(AndFilterList([f for f in filters]))
        self.send_info(f"{[str(f) for f in filters] if len(filters)>1 else str(filters[0])} added to {'exclusions' if exclude else 'breakpoints'}.")

    def remove_filter(self, index:int, exclude:bool = False):
        """
        Remove a filter from the list of breakpoints or exclusions.

        Args:
            index (int): Index of the filter found using list_filters.
            exclude (bool): True to remove from exclusions, False to remove from breakpoints.
        
        Short name: re
        """
        lst = self.debugger.exclusions if exclude else self.debugger.breakpoints
        if index >= len(lst):
            raise DebuggerCommandError(f"Index {index} out of range.")
        fltr = lst.pop(index)
        self.send_info(f"{fltr} removed from {'exclusions' if exclude else 'breakpoints'}.")

    def clear_filters(self):
        """
        Clear all filters from the list of breakpoints and exclusions.
  
        Short name: cl
        """
        self.debugger.exclusions.clear()
        self.debugger.breakpoints.clear()
        self.send_info(f"All filters cleared.")

    def make_filter(self, function:str):
        """
        Add a filtering method. A method which returns a boolean value to indicate that the filter matches or not. When the method returns True, the filter matches and the breakpoint is triggered.

        Args:
            function (str): Name of the filter function.
        
        Short name: fm
        """
        func = self.debugger.context.get_definition(function)
        return FilterFunction(func, debugger=self.debugger)
    
    def or_filters(self, *filters:Filter):
        """
        Add an OR filter.

        Args:
            filters (tuple): The filters to OR.
        
        Short name: f_or
        """
        return OrFilterList([f for f in filters])
    def and_filters(self, *filters:Filter):
        """
        Add an AND filter.

        Args:
            filters (tuple): The filters to AND.
        
        Short name: f_and
        """
        return AndFilterList([f for f in filters])
    
    def by_address(self, address:int):
        """
        Add an address filter.

        Args:
            address (int): Address to filter on.
        
        Short name: fa
        """
        return AddressFilter(address)

    def by_line(self, source_file:str, line_nr:int):
        """
        Set a source line filter.

        Args:
            source_file (str): The source file to filter on.
            line_nr (int): The line number to filter on.
        
        Short name: fline
        """
        address = self.debugger.find_address(source_file, line_nr)
        if address is None:
            raise DebuggerCommandError(f"No address found for {source_file}:{line_nr}.")
        return SourceFilter(address, source_file, line_nr)

    def by_function(self, name:str):
        """
        Specify a function by name in the binary to filter on.

        Args:
            name (str): The name of the function on which you want to filter.
        
        Short name: ffun
        """
        if self.debugger.get_function_address(name) is None:
            raise DebuggerCommandError(f"Function {name} not found.")
        return FunctionFilter(name)

    def by_stream(self, text:str, stream:StreamType=StreamType.stdout):
        """
        Set a stream filter.

        Args:
            text (str): Text that must be in the stream.
            stream (StreamType): Stream to filter (i.e., stdin/stdout/stderr). Default 'stdout'.
        
        Short name: fstr
        """
        return StdStreamFilter(stream.value, text)



from array import array
import codecs
from enum import Enum, auto
import importlib
import os
import inspect
from typing import Literal, cast, get_args

import archinfo
from claripy import List
import re

import claripy
import platform

from dAngr.exceptions import DebuggerCommandError, InvalidArgumentError, ValueError



#create a undefined type
undefined = type("Undefined", (), {})


class Operator(Enum):
    POW = auto()
    MOD = auto()
    MUL = auto()
    DIV = auto()
    FLOORDIV = auto()  
    ADD = auto()
    SUB = auto()
    LSHIFT = auto() 
    RSHIFT = auto()  
    XOR = auto()     
    BITWISE_AND = auto()  
    EQ = auto()
    NEQ = auto()
    GT = auto()
    LT = auto()
    LE = auto()
    GE = auto()
    BITWISE_OR = auto()  
    AND = auto()  
    OR = auto()

    def __repr__(self):
        switch = {
            Operator.POW: "**",
            Operator.MOD: "%",
            Operator.MUL: "*",
            Operator.DIV: "/",
            Operator.FLOORDIV: "//",
            Operator.ADD: "+",
            Operator.SUB: "-",
            Operator.LSHIFT: "<<",
            Operator.RSHIFT: ">>",
            Operator.XOR: "^",
            Operator.BITWISE_AND: "&",
            Operator.BITWISE_OR: "|",
            Operator.EQ: "==",
            Operator.NEQ: "!=",
            Operator.GT: ">",
            Operator.LT: "<",
            Operator.LE: "<=",
            Operator.GE: ">=",
            Operator.AND: "and",
            Operator.OR: "or"
        }
        return switch[self]
    def __str__(self) -> str:
        return self.__repr__()
    @property
    def precedence(self):
        return Operator.operator_precedence(self)
    
    @staticmethod
    def operator_precedence(op):
        switch = {
            Operator.POW : 7,
            Operator.MOD : 6,
            Operator.MUL : 6,
            Operator.DIV : 6,
            Operator.FLOORDIV : 6,  
            Operator.ADD : 5,
            Operator.SUB : 5,
            Operator.LSHIFT : 4,
            Operator.RSHIFT : 4,  
            Operator.XOR : 4, 
            Operator.BITWISE_AND : 3,
            Operator.EQ : 3,
            Operator.NEQ : 3,
            Operator.GT : 3,
            Operator.LT : 3,
            Operator.LE : 3,
            Operator.GE : 3,
            Operator.BITWISE_OR : 2,
            Operator.AND : 2,
            Operator.OR : 1, 
        }
        return switch.get(op, 0)
    # Method resolution based on Python's magic methods
    @staticmethod
    def convert_to_magic_method(op):
        switch = {
            Operator.POW: "__pow__",
            Operator.MOD: "__mod__",
            Operator.MUL: "__mul__",
            Operator.DIV: "__truediv__",  
            Operator.FLOORDIV: "__floordiv__",  
            Operator.ADD: "__add__",
            Operator.SUB: "__sub__",
            Operator.LSHIFT: "__lshift__",  
            Operator.RSHIFT: "__rshift__",  
            Operator.XOR: "__xor__",       
            Operator.BITWISE_AND: "__and__",  
            Operator.BITWISE_OR: "__or__",   
            Operator.EQ: "__eq__",
            Operator.NEQ: "__ne__",  
            Operator.GT: "__gt__",
            Operator.LT: "__lt__",
            Operator.LE: "__le__",
            Operator.GE: "__ge__"
        }
        return switch.get(op, None)
    

class Endness(Enum):
    LE = auto()
    BE = auto()
    DEFAULT = auto()
    MEMORY = auto()
    REGISTER = auto()
    def to_arch_endness(self, project):
        switch = {
            Endness.LE: archinfo.Endness.LE,
            Endness.BE: archinfo.Endness.BE,
            Endness.MEMORY: project.arch.memory_endness,
            Endness.REGISTER: project.arch.register_endness,
            Endness.DEFAULT: archinfo.Endness.BE
        }
        return switch.get(self, archinfo.Endness.BE)
    
    def to_byteorder(self, arch:archinfo.Arch)->Literal["little", "big"]:
        little = 'little'
        big = 'big'
        switch = {
            Endness.LE: little,
            Endness.BE: big,
            Endness.MEMORY: little if arch.memory_endness == archinfo.Endness.LE else big,
            Endness.REGISTER: little if arch.register_endness == archinfo.Endness.LE else big,
            Endness.DEFAULT: big
        }
        en = switch.get(self, big)
        return cast(Literal["little", "big"], en)

class SolverType(Enum):
    UpTo = auto()
    AtLeast = auto()
    Exact = auto()

class DataType(Enum):
    int = auto()
    str = auto()
    bytes = auto()
    bool = auto()
    hex = auto()
    address = auto()
    none = auto()

    def to_type(self)->type:
        switch = {
            DataType.int: int,
            DataType.str: str,
            DataType.bytes: bytes,
            DataType.bool: bool,
            DataType.hex: str,
            DataType.address: int,
            DataType.none: type(None)
        }
        return switch.get(self, type(None))

    def convert(self, value, arch:archinfo.Arch, **kwargs):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        switch = {
            DataType.int: DataType._to_int,
            DataType.str: DataType._to_str,
            DataType.bytes: DataType._to_bytes,
            DataType.bool: DataType._to_bool,
            DataType.hex: DataType._to_hex,
            DataType.address: DataType._to_address
        }
        return switch.get(self, None)(value, arch, **kwargs)

    @staticmethod
    def _to_int(value, arch:archinfo.Arch, endness:Endness=Endness.DEFAULT):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        if type(value) == int:
            return value
        elif type(value) == str:
            if value.startswith('0x'):
                return int(value, 16)
            return int(value, 0)
        elif type(value) == bytes:
            return int.from_bytes(value, byteorder=endness.to_byteorder(arch))
        elif type(value) == bool:
            return int(value)
        raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', or 'bytes'.")
    
    @staticmethod
    def _to_address(value, arch:archinfo.Arch, endness:Endness=Endness.DEFAULT):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        if type(value) == int:
            v = value
        elif type(value) == str:
            v = int(value, 0)
        elif type(value) == bytes:
            v = int.from_bytes(value, byteorder=endness.to_byteorder(arch))
        elif type(value) == bool:
            v = int(value)
        else:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', or 'bytes'.")
        return hex(v)
    
    @staticmethod
    def _to_hex( value, arch:archinfo.Arch, endness:Endness=Endness.DEFAULT):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        if type(value) == int:
            return hex(value)
        elif type(value) == str:
            return DataType._to_bytes(value, arch, endness).hex(":")
        elif type(value) == bytes:
            return value.hex(":")
        elif type(value) == bool:
            return hex(int(value))
        raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', or 'bytes'.")

    @staticmethod
    def _to_str(value, arch:archinfo.Arch):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        if type(value) == int:
            return str(value)
        elif type(value) == bytes:
            return value.decode('utf-8', errors='replace')
        elif type(value) == str:
            return value
        elif type(value) == bool:
            return str(value)
        raise DebuggerCommandError(f"Type not supported. Use 'int' or 'bytes'.")

    @staticmethod
    def _to_bool(value, arch:archinfo.Arch, endness:Endness=Endness.DEFAULT):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        if type(value) == int:
            return bool(value)
        elif type(value) == str:
            return value.lower() in ['true', '1']
        elif type(value) == bytes:
            return bool(int.from_bytes(value, byteorder=endness.to_byteorder(arch)))
        elif type(value) == bool:
            return value
        raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', or 'bytes'.")

    @staticmethod
    def _to_bytes(value, arch:archinfo.Arch, endness:Endness=Endness.DEFAULT):
        if not type(value) in [int, str, bytes, bool]:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', 'bytes', or 'bool'.")
        if type(value) == int:
            # Account for endianness when storing integers
            byte_value = value.to_bytes(arch.bytes, byteorder=endness.to_byteorder(arch))
        elif type(value) == str:
            byte_value = value.encode('utf-8')
        elif type(value) == bytes:
            byte_value = value
        elif type(value) == bool:
            byte_value = int(value).to_bytes(1, byteorder=endness.to_byteorder(arch))
        else:
            raise DebuggerCommandError(f"Type not supported. Use 'int', 'str', or 'bytes'.")
        return byte_value

class StreamType(Enum):
    stdin = 0
    stdout = 1
    stderr = 2
    
class ObjectStore(Enum):
    mem = auto()
    sym = auto()
    reg = auto()
    io = auto()



Constraint = claripy.ast.Bool
SymBitVector = claripy.ast.BV
class Variable:
    @property
    def value(self):
        pass
    @value.setter
    def value(self, value):
        pass

AngrValueType = SymBitVector | int | str | bytes | bool
AngrObjectType = AngrValueType | Variable
AngrType = AngrValueType | AngrObjectType
AngrCompoundType = dict[str, AngrValueType] | list[AngrValueType]
AngrExtendedType = Variable | AngrCompoundType

class Variable:
    def __init__(self, name:str, value:AngrExtendedType):
        self.name = name
        assert not isinstance(value, Variable)
        self._value = value

    @property
    def value(self) ->AngrExtendedType:   
        return self._value
    @value.setter
    def value(self, value:AngrExtendedType):
        assert isinstance(value, AngrValueType) or isinstance(value, list) or isinstance(value, dict), "Invalid Variable Type"
        self._value = value
    
    def __repr__(self):
        return f"{self.name}={self._value}"


def is_indexable(obj):
    return hasattr(obj, '__getitem__')

def is_iterable(obj):
    return hasattr(obj, '__iter__') or isinstance(obj, array)

def str_to_type(dtype:str):
    #convert string dtype to typings type
    tp = None
    if dtype == '':
        tp = None
    elif dtype == "int":
        tp = int
    elif dtype == "str":
        tp = str
    elif dtype == "bytes":
        tp = bytes
    elif dtype == "bool":
        tp = bool
    elif dtype == "double":
        tp = float
    elif dtype == "hex":
        tp = int
    elif dtype == "tuple":
        tp = tuple
    elif dtype == "list":
        tp = list
    elif dtype == "dict":
        tp = dict
    elif dtype == "set":
        tp = set
    elif dtype == 'array':
        tp = array
    else:
        try:
            from dAngr.cli.grammar.expressions import ReferenceObject,VariableRef, SymbolicValue, Register, Property, IndexedProperty
            import angr
            tp = eval(dtype)
        except:
            raise ValueError(f"Invalid data type {dtype}")
    return tp

def get_local_arch():
    # Detect platform architecture
    machine = platform.machine()
    arch_map = {
        'x86_64': archinfo.ArchAMD64,
        'i386': archinfo.ArchX86,
        'arm': archinfo.ArchARM,
        'aarch64': archinfo.ArchAArch64,
        'mips': archinfo.ArchMIPS32,
        'mips64': archinfo.ArchMIPS64,
    }
    if machine in arch_map:
        return arch_map[machine]()
    else:
        raise ValueError(f"Unsupported architecture: {machine}")

def check_signature_matches(func, o, args, kwargs):
    # Get the function's signature
    signature = inspect.signature(func)
    
    try:
        # Bind the provided arguments to the function's signature
        bound_args = signature.bind(o, *args, **kwargs)
        bound_args.apply_defaults()  # Apply default values if any
    except TypeError as e:
        raise InvalidArgumentError(str(e))
    
    # Optionally: Perform type checking if the function has type hints

    for param_name, param_value in bound_args.arguments.items():
            p = signature.parameters.get(param_name)
            if p is None:
                raise InvalidArgumentError(f"Invalid argument '{param_name}', not found in signature")
            if p.name == "self":
                continue
            if p is None:
                raise InvalidArgumentError(f"Invalid argument '{param_name}', not found in signature")
            
            expected_type = p.annotation
            if inspect.isclass(expected_type) and  issubclass(expected_type ,Enum) and isinstance(param_value, str):
                #get the value of the expected enum type given the str
                param_value = expected_type[param_value]
            if p.kind == inspect.Parameter.VAR_POSITIONAL:
                    if not isinstance(param_value, tuple):
                        raise InvalidArgumentError(f"Argument '{param_name}' should be of type {expected_type.__name__ if isinstance(expected_type, type) else expected_type}, got {type(param_value).__name__}")
                    for t in param_value:
                        if expected_type != inspect._empty and  not isinstance(t, expected_type):
                            raise InvalidArgumentError(f"Argument '{param_name}' should be of type {expected_type.__name__ if isinstance(expected_type, type) else expected_type}, got {type(param_value).__name__}")
            elif p.kind == inspect.Parameter.VAR_KEYWORD:
                if not isinstance(param_value, dict):
                    raise InvalidArgumentError(f"Argument '{param_name}' should be of type {expected_type.__name__ if isinstance(expected_type, type) else expected_type}, got {type(param_value).__name__}")
                for k,v in param_value.items():
                    if expected_type != inspect._empty and  not isinstance(v, expected_type):
                        raise InvalidArgumentError(f"Argument '{param_name}' should be of type {expected_type.__name__ if isinstance(expected_type, type) else expected_type}, got {type(param_value).__name__}")
            elif expected_type != inspect._empty and not isinstance(param_value, expected_type):
                raise InvalidArgumentError(f"Argument '{param_name}' should be of type {expected_type.__name__ if isinstance(expected_type, type) else expected_type}, got {type(param_value).__name__}")
    
def parse_binary_string(binary_string_text):
    # Strip the `b'` prefix and trailing `'`
    if (binary_string_text.startswith("b'") and binary_string_text.endswith("'")) or (binary_string_text.startswith("b\"") and binary_string_text.endswith("\"")):
        binary_string_text = binary_string_text[2:-1]
    else:
        raise ValueError("Invalid binary string format")
    
    # Handle escape sequences (e.g., \0, \n, \\)
    parsed_string = codecs.decode(binary_string_text, 'unicode_escape')

    # Convert to bytes using 'latin1' to preserve the byte-to-byte mapping
    binary_data = parsed_string.encode('latin1')

    return binary_data

def convert_argument(arg_type: type, arg_value: str):
    try:
        # Handle Enums
        if isinstance(arg_type, type) and issubclass(arg_type, Enum):
            return arg_type[arg_value]

        # Handle Union types
        types = [arg_type]
        if members := get_union_members(arg_type):
            types = members
        if bool in types and arg_value.lower() in ['true', 'false']:
            return arg_value.lower() == 'true'
        if bytes in types:
            if (arg_value.startswith('b"') and arg_value.endswith('"')) or (arg_value.startswith("b'") and arg_value.endswith("'")):
                return bytes(arg_value[2:-1], 'utf-8')
        if int in types:
            if arg_value.startswith('0x'):
                if "^" in arg_value:
                    arg_value,_ = arg_value.split("^")
                    return int(arg_value, 16) ^ int(arg_value.split("^")[1], 16)
                return int(arg_value, 16)
            if arg_value.isnumeric():
                return int(arg_value)
            if arg_value.startswith('-') and arg_value[1:].isnumeric():
                return - int(arg_value[1:])
        if str in types:
            if arg_value.startswith(('\'', '"')) and arg_value.endswith(('\'', '"')):
                return arg_value[1:-1]
            return arg_value

        # If no type matched
        raise InvalidArgumentError(f"Failed to convert argument to any of the expected types {types} from '{arg_value}'")
    except ValueError:
        raise InvalidArgumentError(f"Failed to convert argument to type '{arg_type}' from '{arg_value}'")

def str_to_address(address:str):
    if address.startswith("0x"):
        return int(address, 16)
    return int(address)



def remove_xml_tags(text):
    # Use a regular expression to match and remove all tags
    clean_text = re.sub(r'<[^>]+>', '', text)
    return clean_text

def get_union_members(union_type):
    return get_args(union_type)

def parse_arguments(input:str, splitter):
    # Construct a regex pattern from the command's argument specifications
        # string arguments can be a signle word or words in double quotes
        # parse user_input:
        # 1. split by spaces
        # 2. if a word starts with double quote, join all words until the next double quote
        tokens:List[str] = [t.strip() for t in input.split(splitter)]
        if len(tokens) == 1 and not tokens[0]:
            tokens = []
        parsed_args = []
        i = 0
        while i < len(tokens):
            if tokens[i].startswith(('\'','"', 'b"','b\'')):
                t = tokens[i][0]
                j = i
                while j<len(tokens) and  not tokens[j].endswith(t):
                    j += 1
                parsed_args.append(splitter.join(tokens[i:j+1]))
                i = j
            else:
                parsed_args.append(tokens[i])
            i += 1
        return parsed_args

def get_python_classes_in_folder(folder_path):
    python_classes = []

    # Traverse the directory structure recursively
    for root, dirs, files in os.walk(folder_path):
        for file in files:
            if file.endswith('.py'):
                file_path = os.path.join(root, file)
                classes = get_classes_from_file(file_path)
                python_classes.extend(classes)

    return python_classes

def get_classes_from_file(file_path):
    classes = []

    # Attempt to import the module dynamically
    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path) # type: ignore
    if spec is not None:
        module = importlib.util.module_from_spec(spec) # type: ignore
        try:
            spec.loader.exec_module(module)

            # Iterate over the members of the module
            for name, obj in inspect.getmembers(module):
                # Check if the member is a class
                if inspect.isclass(obj):
                    classes.append(obj)
                
        except:
            pass
    return classes


def remove_ansi_escape_codes(text):
    ansi_escape = re.compile(r'(?:\x1B[@-_]|[\\x1B\x9B])[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', text)





# Generated from /workspaces/dAngr/src/dAngr/cli/grammar/ranges.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .rangesParser import rangesParser
else:
    from rangesParser import rangesParser

# This class defines a complete generic visitor for a parse tree produced by rangesParser.

class rangesVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by rangesParser#expression.
    def visitExpression(self, ctx:rangesParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#range.
    def visitRange(self, ctx:rangesParser.RangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#dangr_range.
    def visitDangr_range(self, ctx:rangesParser.Dangr_rangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#bash_range.
    def visitBash_range(self, ctx:rangesParser.Bash_rangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#python_range.
    def visitPython_range(self, ctx:rangesParser.Python_rangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#bash_content.
    def visitBash_content(self, ctx:rangesParser.Bash_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#py_content.
    def visitPy_content(self, ctx:rangesParser.Py_contentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#anything.
    def visitAnything(self, ctx:rangesParser.AnythingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by rangesParser#symbol.
    def visitSymbol(self, ctx:rangesParser.SymbolContext):
        return self.visitChildren(ctx)



del rangesParser



from angr import SimStatePlugin


class StdTracker(SimStatePlugin):
    def __init__(self, stream:int=1):
        super(StdTracker, self).__init__()
        self.std_mark = 0
        self.prev_std_mark = 0
        self.stream = stream

    def set_mark(self):
        self.prev_std_mark = self.std_mark
        self.std_mark = len(self.state.posix.dumps(self.stream))

    def get_new_data(self)->bytes:
        start_len = self.std_mark
        full_output = self.state.posix.dumps(self.stream)
        new_output = full_output[start_len:]
        self.set_mark()
        return new_output
    
    def get_new_string(self)->str:
        return self.get_new_data().decode()
    
    def get_prev_string(self)->str:
        if self.prev_std_mark == self.std_mark:
            return ''
        start_len = self.prev_std_mark
        full_output = self.state.posix.dumps(self.stream)
        return full_output[start_len:].decode('utf-8')

    @SimStatePlugin.memo
    def copy(self, _memo)-> SimStatePlugin: # type: ignore
        # This method is used by angr to copy the plugin when forking states
        new_plugin = StdTracker(stream=self.stream)
        new_plugin.std_mark = self.std_mark
        return new_plugin




from typing import cast
from antlr4 import TerminalNode
from dAngr.exceptions import ParseError
from dAngr.cli.grammar.antlr.dAngrParser import dAngrParser
from dAngr.cli.grammar.antlr.dAngrVisitor import dAngrVisitor
from dAngr.cli.grammar.statements import Assignment,  Statement
from dAngr.cli.grammar.control_flow import IfThenElse, WhileLoop, ForLoop
from dAngr.cli.grammar.script import Script, Body
from dAngr.cli.grammar.definitions import ArgumentSpec, CustomFunctionDefinition
from dAngr.cli.grammar.expressions import BREAK, CONTINUE, BASECommand, DangrCommand, Dictionary, IfConstraint, Inclusion, Invert, Listing, Memory, Negate, Operator, PythonCommand, BashCommand, Comparison, Literal, Property, IndexedProperty, Range, ReferenceObject, Slice, StateObject, VariableRef
from dAngr.utils.utils import parse_binary_string


class dAngrVisitor_(dAngrVisitor):
    def __init__(self, debugger):
        self.debugger = debugger
        self.operators = {
            "**": Operator.POW,
            "%": Operator.MOD,
            "*": Operator.MUL,
            "/": Operator.DIV,
            "//": Operator.FLOORDIV,
            "+": Operator.ADD,
            "-": Operator.SUB,
            "<<": Operator.LSHIFT,
            ">>": Operator.RSHIFT,
            "&": Operator.BITWISE_AND,
            "|": Operator.BITWISE_OR,
            "<": Operator.LT,
            "<=": Operator.LE,
            ">": Operator.GT,
            ">=": Operator.GE,
            "==": Operator.EQ,
            "!=": Operator.NEQ,
            "^": Operator.XOR,
            "&&": Operator.AND,
            "||": Operator.OR
        }
    def _replace_text(self, text:str):
        return text.replace("\\n", "\n").replace("\\t", "\t").replace("\\r", "\r")
    
    def getOperator(self, op):
        if op in self.operators:
            return self.operators[op]
        else:
            raise ParseError(f"Operator {op} not supported")

    def visitScript(self, ctx: dAngrParser.ScriptContext):
        if ctx.QMARK() or ctx.HELP():
            args = []
            if ctx.identifier():
                cmd = self.visit(ctx.identifier())
                args = [cmd]
            return Script([DangrCommand("help",None, *args)],[]) # type: ignore
        else:
            statements = [self.visit(s) for s in ctx.statement()] if ctx.statement() else []
            definitions = [self.visit(c) for c in ctx.function_def()]
            statements = Statement.flatten(statements)
            return Script(statements, definitions)
    
    def visitStatement(self, ctx: dAngrParser.StatementContext):
        if ctx.assignment():
            return self.visit(ctx.assignment())
        elif ctx.control_flow():
            return self.visit(ctx.control_flow())
        elif ctx.expression():
            return self.visit(ctx.expression())
        elif ctx.ext_command():
            return self.visit(ctx.ext_command())
        elif ctx.static_var(): # static variable
            return VariableRef(self.visit(ctx.static_var().identifier()),True)
        raise ParseError(f"Invalid statement {ctx.getText()}")
    
    # def visitExpression(self, ctx: dAngrParser.ExpressionContext):
    #     if ctx.identifier():
    #         start = 0
    #         if ctx.DOT():
    #             package = ctx.identifier(0).getText()
    #             cmd =  ctx.identifier(1).getText()
    #             start = 3
    #         else:
    #             package = None
    #             cmd = ctx.identifier(0).getText()
    #             start = 1
    #         args = []
    #         kwargs  = {}
    #         if ctx.expression_part():
    #             children = ctx.children[start:]
    #             for i in range(0, len(children)):
    #                 #check if c is a terminalnode drop it
    #                 c = children[i]
    #                 if isinstance(c, TerminalNode):
    #                     continue

    #                 #if c is an identifier, it is a named argument
    #                 if isinstance(c, dAngrParser.IdentifierContext):
    #                     name = c.getText()
    #                     for j in range(i+1, len(children)):
    #                         if isinstance(children[j], TerminalNode):
    #                             continue
    #                         if isinstance(children[j], dAngrParser.IdentifierContext):
    #                             name = children[j].getText()
    #                         elif isinstance(children[j], dAngrParser.Expression_partContext):
    #                             kwargs[name] = self.visit(children[j])
    #                     break
    #                 else:
    #                     assert kwargs == {}
    #                     args.append(self.visit(c))
    #         return DangrCommand(cmd, package, *args, **kwargs)
    #     elif ctx.constraint():
    #         return self.visit(ctx.constraint())
    #     elif ctx.expression_part():
    #         return self.visit(ctx.expression_part(0))
    #     else:
    #         raise ParseError(f"Invalid expression {ctx.getText()}")
    
    # def visitConstraint(self, ctx: dAngrParser.ConstraintContext):
    #     if ctx.CIF(): # if constraint
    #         iif = self.visit(ctx.condition().expression())
    #         cthen = self.visit(ctx.expression_part(0))
    #         celse = self.visit(ctx.expression_part(1))
    #         return IfConstraint(iif, cthen, celse)
    #     else:
    #         raise ParseError(f"Invalid constraint {ctx.getText()}")

    def visitExpression(self, ctx: dAngrParser.ExpressionContext):
        if ctx.identifier():
            start = 0
            if ctx.DOT():
                package = ctx.identifier(0).getText()
                cmd =  ctx.identifier(1).getText()
                start = 3
            else:
                package = None
                cmd = ctx.identifier(0).getText()
                start = 1
            args = []
            kwargs  = {}
            if ctx.expression_part():
                children = ctx.children[start:]
                for i in range(0, len(children)):
                    #check if c is a terminalnode drop it
                    c = children[i]
                    if isinstance(c, TerminalNode):
                        continue

                    #if c is an identifier, it is a named argument
                    if isinstance(c, dAngrParser.IdentifierContext):
                        name = c.getText()
                        for j in range(i+1, len(children)):
                            if isinstance(children[j], TerminalNode):
                                continue
                            if isinstance(children[j], dAngrParser.IdentifierContext):
                                name = children[j].getText()
                            elif isinstance(children[j], dAngrParser.Expression_partContext):
                                kwargs[name] = self.visit(children[j])
                        break
                    else:
                        assert kwargs == {}
                        args.append(self.visit(c))
            from dAngr.cli.command_line_debugger import DEBUGGER_COMMANDS
            if not(self.debugger and self.debugger.context):
                if cmd in DEBUGGER_COMMANDS:
                    return DangrCommand(cmd, None, *args, **kwargs)
            if self.debugger and self.debugger.context and self.debugger.context.find_function(package, cmd):
                return DangrCommand(cmd, package, *args, **kwargs)
            else:
                return VariableRef(Literal(cmd))
        else:
            return self.visit(ctx.expression_part(0))
    def visitExpressionIf(self, ctx: dAngrParser.ExpressionIfContext):
        iif = self.visit(ctx.condition())
        cthen = self.visit(ctx.expression_part(0))
        celse = self.visit(ctx.expression_part(1))
        return IfConstraint(iif, cthen, celse)      
    def visitExpressionIn(self, ctx: dAngrParser.ExpressionInContext):
        return Inclusion(self.visit(ctx.expression_part(0)), self.visit(ctx.expression_part(1)), ctx.NOT_IN() is not None)
    def visitExpressionAlt(self, ctx: dAngrParser.ExpressionAltContext):
        return self.visit(ctx.range_())
    def visitExpressionParenthesis(self, ctx: dAngrParser.ExpressionParenthesisContext):
        return self.visit(ctx.expression())
    def visitExpressionBool(self, ctx: dAngrParser.ExpressionBoolContext):
        return Literal(ctx.BOOL().getText() == "True")
    def visitExpressionObjectContext(self, ctx: dAngrParser.ExpressionObjectContext):
        return self.visit(ctx.object_())
    def visitExpressionRange(self, ctx: dAngrParser.ExpressionRangeContext):
        start = self.visit(ctx.expression_part(0))
        if len(ctx.expression_part())==1:
            return Range(start)
        end = self.visit(ctx.expression_part(1))
        if len(ctx.expression_part())==2:
            return Range(start, end)
        step = self.visit(ctx.expression_part(2))
        return Range(start, self.visit(ctx.expression_part(1)), step)
    def visitExpressionOperation(self, ctx: dAngrParser.ExpressionOperationContext):
        lhs = self.visit(ctx.object_())
        op = self.getOperator(ctx.operation().getText())
        rhs = self.visit(ctx.expression_part())
        return Comparison(lhs, op, rhs)
    def visitExpressionNot(self, ctx: dAngrParser.ExpressionNotContext):
        return Invert(self.visit(ctx.expression_part()))
    
    def visitExpressionReference(self, ctx: dAngrParser.ExpressionReferenceContext):
        return self.visit(ctx.reference())
    
    def visitAssignment(self, ctx: dAngrParser.AssignmentContext):
        if ctx.static_var():
            var = VariableRef(self.visit(ctx.static_var().identifier()),True)
        else:
            var = self.visit(ctx.object_())
        # check if variable name does not match a short name of a command
        if isinstance(var, VariableRef):
            name = var.name
            if isinstance(name, Literal):
                v = name.get_value(None)
                if not (self.debugger and self.debugger.context):
                    from dAngr.cli.command_line_debugger import DEBUGGER_COMMANDS
                    if v in DEBUGGER_COMMANDS:
                        raise ParseError(f"Variable name {v} is a command name") 
                else:
                    from dAngr.cli.command_line_debugger import CommandLineDebugger
                    dbg = cast(CommandLineDebugger, self.debugger)
                    if f:=dbg.context.find_function(None, v):
                        raise ParseError(f"Variable name {v} is a command name for {f.name}")
        val = self.visit(ctx.expression())
        return Assignment(var, val)
    
    def visitExt_command(self, ctx: dAngrParser.Ext_commandContext):
        if ctx.BANG():
            args = self.visit(ctx.py_basic_content())
            return PythonCommand(*args[0],**args[1])
        elif ctx.AMP():
            return self.visit(ctx.expression())
        elif ctx.DOLLAR():
            return self.visit(ctx.bash_content())
        raise ParseError(f"Invalid command {ctx.getText()}")
    
    def visitControl_flow(self, ctx: dAngrParser.Control_flowContext):
        if ctx.IF():
            return IfThenElse(self.visit(ctx.condition()), self.visit(ctx.body()), self.visit(ctx.else_().body()) if ctx.else_() else Body([]))
        elif ctx.WHILE():
            return WhileLoop(self.visit(ctx.condition()), self.visit(ctx.body()))
        elif ctx.FOR():
            index = None
            item = None
            if len(ctx.identifier())==2:
                index = VariableRef(self.visit(ctx.identifier(0)))
                item = VariableRef(self.visit(ctx.identifier(1)))
            else:
                item = VariableRef(self.visit(ctx.identifier(0)))
            return ForLoop(self.visit(ctx.iterable()), self.visit(ctx.body()),item, index)
        raise ParseError("Invalid control flow")    


    def visitFunction_def(self, ctx: dAngrParser.Function_defContext):
        name = ctx.identifier().getText()
        args = []
        if ctx.parameters():
            args = self.visit(ctx.parameters())
        body = self.visit(ctx.body())
        return CustomFunctionDefinition(name, args, body)
    
     
    def visitBody(self, ctx: dAngrParser.BodyContext):
        statements = [self.visit(s) for s in ctx.fstatement()] if ctx.fstatement() else []
        return Body(Statement.flatten(statements))
    
    def visitFstatement(self, ctx: dAngrParser.FstatementContext):
        if ctx.BREAK():
            return BREAK
        elif ctx.CONTINUE():
            return CONTINUE
        elif ctx.RETURN():
            return BASECommand("return", self.visit(ctx.expression()) if ctx.expression() else None)
        elif ctx.expression():
            return self.visit(ctx.expression()) # TODO: deal with return
        elif ctx.statement():
            return self.visit(ctx.statement())

    def visitIterable(self, ctx: dAngrParser.IterableContext):
        # if not ctx.LPAREN():
        return self.visit(ctx.expression())
        # else:
        
    def visitParameters(self, ctx: dAngrParser.ParametersContext):
        return [ArgumentSpec(p.getText())for p in ctx.identifier()]

    def visitCondition(self, ctx: dAngrParser.ConditionContext):
        return self.visit(ctx.expression())
    
#OBJECTS
    def visitIDObject(self, ctx: dAngrParser.IDObjectContext):
        v = VariableRef(self.visit(ctx.identifier()))
        if ctx.BANG():
            return DangrCommand("evaluate", None, v)
        return v
    def visitNumericObject(self, ctx: dAngrParser.NumericObjectContext):
        if ctx.DASH():
            return Negate(self.visit(ctx.numeric()))
        return self.visit(ctx.numeric())
    
    def visitBoolObject(self, ctx: dAngrParser.BoolObjectContext):
        return Literal(ctx.BOOL().getText() == "True")
    def visitReferenceObject(self, ctx: dAngrParser.ReferenceObjectContext):
        return self.visit(ctx.reference())
    def visitPropertyObject(self, ctx: dAngrParser.PropertyObjectContext):
        o = self.visit(ctx.object_())
        return Property(o, ctx.identifier().getText())
    
    def visitIndexedPropertyObject(self, ctx: dAngrParser.IndexedPropertyObjectContext):
        o = self.visit(ctx.object_())
        index = self.visit(ctx.index())
        return IndexedProperty(o, index)
    def visitSliceStartEndObject(self, ctx: dAngrParser.SliceStartEndObjectContext):
        o = self.visit(ctx.object_())
        start = self.visit(ctx.index(0))
        if len(ctx.index())>1:
            end = self.visit(ctx.index(1))
        else:
            end = -1
        return Slice(o, start, end)
    def visitSlideStartLengthObject(self, ctx: dAngrParser.SlideStartLengthObjectContext):
        o = self.visit(ctx.object_())
        start = self.visit(ctx.index(0))
        length = self.visit(ctx.index(1))
        return Slice(o, start, Comparison(start, Operator.ADD, length))
    
    def visitListObject(self, ctx: dAngrParser.ListObjectContext):
        objs = [self.visit(o) for o in ctx.object_()]
        return Listing(objs)
    def visitDictionaryObject(self, ctx: dAngrParser.DictionaryObjectContext):
        l = len(ctx.object_())
        d = {}
        for i in range(l):
            d[ctx.STRING(i).getText().strip("'\"")] = self.visit(ctx.object_(i))
        return Dictionary(d)
    def visitStringObject(self, ctx: dAngrParser.StringObjectContext):
        return Literal(self._replace_text(ctx.STRING().getText()[1:-1]))
    def visitBinaryStringObject(self, ctx: dAngrParser.BinaryStringObjectContext):
        return Literal(parse_binary_string(ctx.BINARY_STRING().getText()))
#END OBJECTS
       
    def visitIndex(self, ctx: dAngrParser.IndexContext):
        if ctx.DASH():
            return Negate(self.visit(ctx.expression()))
        else:    
            return self.visit(ctx.expression())
    
    def visitIdentifier(self, ctx: dAngrParser.IdentifierContext):
        return Literal(ctx.getText())
    
    def visitNumeric(self, ctx: dAngrParser.NumericContext):
        return Literal(int(ctx.NUMBERS().getText()) if ctx.NUMBERS() else int(ctx.HEX_NUMBERS().getText(), 16))
    
   
    def _isSTRING(self, ctx):
        if isinstance(ctx, TerminalNode):
            return False
        c =ctx.getTokens(dAngrParser.STRING)
        if len(c) == 1:
            return True
        return (self._isSTRING(ctx.children[0]) if len(ctx.children) == 1 else False)

    def visitPy_basic_content(self, ctx: dAngrParser.Py_basic_contentContext):
        name = ctx.identifier().getText()
        args = []
        kwargs = {}
        for c in ctx.children[1:]:
            if isinstance(c, dAngrParser.Py_contentContext):
                py = self.visit(c)
                args.extend(py.cmds)
                kwargs.update(py.kwargs)
            else:
                args.append(Literal(c.getText()))
        return [name]+args, kwargs
    def visitPy_content(self, ctx: dAngrParser.Py_contentContext):

        args = []
        kwargs = {}
        for c in ctx.children:
            if isinstance(c, dAngrParser.RangeContext):
                args.append(self.visit(c))
            elif isinstance(c, dAngrParser.AnythingContext):
                args.append(self.visit(c))
            elif isinstance(c, dAngrParser.ReferenceContext):
                args.append(self.visit(c))
            elif isinstance(c, dAngrParser.Py_contentContext):
                py:PythonCommand = self.visit(c)
                args.extend(py.cmds)
                kwargs.update(py.kwargs)
            else:
                args.append(Literal(c.getText()))
        return PythonCommand(*args)
    
    def visitReference(self, ctx: dAngrParser.ReferenceContext):
        if ctx.MEM_DB():
            size = None
            if len(ctx.index()) == 2:
                size = self.visit(ctx.index(1))
            m = Memory(self.visit(ctx.index(0)), size)
            if ctx.BANG():
                return DangrCommand("evaluate", None, m)
            return m
        elif ctx.VARS_DB():
            r = ReferenceObject.createNamedObject(ctx.VARS_DB().getText(), self.visit(ctx.identifier()))
            if ctx.BANG():
                return DangrCommand("evaluate", None, r)
            return r
        elif ctx.REG_DB():
            r = ReferenceObject.createNamedObject(ctx.REG_DB().getText(), self.visit(ctx.identifier()))
            if ctx.BANG():
                return DangrCommand("evaluate", None, r)
            return r
        elif ctx.SYM_DB():
            r = ReferenceObject.createNamedObject(ctx.SYM_DB().getText(), self.visit(ctx.identifier()))
            if ctx.BANG():
                return DangrCommand("evaluate", None, r)
            return r
        elif ctx.STATE():
            return StateObject()
        else:
            raise ParseError(f"Invalid reference {ctx.getText()}")
        
    
    def visitBash_content(self, ctx: dAngrParser.Bash_contentContext):
        args = []
        for c in ctx.children:
            if isinstance(c, dAngrParser.RangeContext):
                args.append(self.visit(c))
            elif isinstance(c, dAngrParser.AnythingContext):
                args.append(self.visit(c))
            elif isinstance(c, dAngrParser.ReferenceContext):
                args.append(self.visit(c))
            elif isinstance(c, dAngrParser.Bash_contentContext):
                ba:BashCommand = self.visit(c)
                args.extend(ba.cmds)
            else:
                args.append(Literal(c.getText()))
        return BashCommand(*args)
    
    def visitSymbol(self, ctx: dAngrParser.SymbolContext):
        return Literal(ctx.getText())
    def visitAnything(self, ctx: dAngrParser.AnythingContext):
        return Literal(ctx.getText())

    def visitRange(self, ctx: dAngrParser.RangeContext):
        if ctx.bash_range():
            return self.visit(ctx.bash_range().bash_content())
        elif ctx.python_range():
            return self.visit(ctx.python_range().py_content())
        else:
            return self.visit(ctx.dangr_range().expression())


from array import array

from dAngr.cli.debugger_commands.base import BaseCommand
from dAngr.exceptions import DebuggerCommandError
from dAngr.utils.utils import DataType, AngrType, AngrValueType, Endness

from dAngr.utils.loggers import get_logger
log = get_logger(__name__)

class ToolCommands(BaseCommand):
    def __init__(self, debugger_core):
        super().__init__(debugger_core)

    def python(self, code:str):
        """
        Execute a python code.

        Args:
            code (str): Python code to execute.

        Short name: !
        """
        try:
            #execute python code and send the result
            return exec(code)
        except Exception as e:
            self.send_error(f"Error: {e}")

    def bash(self, command:str):
        """
        Execute a command in shell.

        Args:
            command (str): Command to execute in shell. May contain spaces.

        Short name: %
        """
        # execute system command and return output
        import subprocess
        try:
            output = subprocess.check_output(command, shell=True)
            return output.decode()
        except subprocess.CalledProcessError as e:
            pass # error already output
        except Exception as e:
            self.send_error(f"Error: {e}")
    
    def get_value(self, ref:str|AngrType):
        """
        Get the value of a variable or symbol.

        Args:
            ref (str|AngrType): Name/ref of the variable or symbol.

        Short name: gv
        """
        return self.get_angr_value(ref)

    def cast_to(self, value:AngrType, dtype:DataType):
        """
        Convert value to a specific data type.

        Args:
            value (AngrType): Value or reference to value to convert.
            dtype (DataType): Data type to convert to.
        
        Short name: to
        
        """
        value = self.get_angr_value(value)
        assert isinstance(value, AngrType), f"Invalid value type {type(value)}"
        return self.debugger.cast_to(value, dtype)
    
    def to_bytes(self, value:AngrType):
        """
        Convert ref to bytes based on current state.

        Args:
            value (AngrType): value or reference to the object. If str then we check either a symbol name or a variable name.
        
        Short name: ctb
        
        """
        return self.cast_to(value, DataType.bytes)
    
    def to_int(self, value:AngrType, endness:Endness=Endness.DEFAULT):
        """
        Solve and get concrete symbol value as int based on current state.

        Args:
            value (AngrType): value or reference to the object. If str then we check either a symbol name or a variable name.
            endness (Endness): Endianness of the value. Default is BE.
        
        Short name: cti
        
        """
        value = self.get_angr_value(value)
        assert isinstance(value, AngrType), f"Invalid value type {type(value)}"
        return self.debugger.cast_to(value, DataType.int, endness=endness)
    
    def to_str(self, value:AngrValueType):
        """
        Convert symbol to a str.

        Args:
            value (AngrValueType): value or reference to the object. If str then we check either a symbol name or a variable name.
        
        Short name: cts
        
        """
        return self.cast_to(value, DataType.str)
    def to_hex(self, value:AngrValueType):
        """
        Convert symbol to a hex representation.

        Args:
            value (AngrValueType): value or reference to the object. If str then we check either a symbol name or a variable name.
        
        Short name: cth
        
        """
        return self.cast_to(value, DataType.hex)
    
    def to_bool(self, value:AngrValueType):
        """
        Solve and get concrete symbol value as bool based on current state.

        Args:
            value (AngrValueType): value or reference to the object. If str then we check either a symbol name or a variable name.
        
        Short name: ctB
        
        """
        return self.cast_to(value, DataType.bool)
    def len(self, value:str|list|array|bytes):
        """
        Get the length of the value object

        Args:
            value (str|list|array|bytes): Value object to get the length of.

        """
        return len(value) # type: ignore
    
    def strip(self, value:str|bytes, chars:str|bytes|None = None):
        """
        Strip the head and tail of the string or bytes

        Args:
            value (str|bytes): Value to strip.
            chars (str|bytes|None): chars to strip. Default None = empty string. 

        """

        #extract \xDD chars

        v = self.debugger.cast_to(value, DataType.bytes) if isinstance(value,str) else value
        c = self.debugger.cast_to(chars, DataType.bytes) if isinstance(chars, str) else chars
        if v and isinstance(v, bytes):
            return v.strip(c)
        else:
            raise DebuggerCommandError("Failed to convert value")
        
    def lstrip(self, value:str|bytes, chars:str|bytes|None = None):
        """
        Strip the head of the string or bytes

        Args:
            value (str|bytes): Value to strip.
            chars (str|bytes|None): chars to strip. Default None = empty string. 

        """

        #extract \xDD chars

        v = self.debugger.cast_to(value, DataType.bytes) if isinstance(value,str) else value
        c = self.debugger.cast_to(chars, DataType.bytes) if isinstance(chars, str) else chars
        if v and isinstance(v, bytes):
            return v.lstrip(c)
        else:
            raise DebuggerCommandError("Failed to convert value")
    
    def rstrip(self, value:str|bytes, chars:str|bytes|None = None):
        """
        Strip the tail of the string or bytes

        Args:
            value (str|bytes): Value to strip.
            chars (str|bytes|None): chars to strip. Default None = empty string. 

        """

        #extract \xDD chars

        v = self.debugger.cast_to(value, DataType.bytes) if isinstance(value,str) else value
        c = self.debugger.cast_to(chars, DataType.bytes) if isinstance(chars, str) else chars
        if v and isinstance(v, bytes):
            return v.rstrip(c)
        else:
            raise DebuggerCommandError("Failed to convert value")
        
    def append(self, value:list, object):
        """
        Append the tail of the value

        Args:
            value (list): Value append to.
            object (): object to append to value

        """
        if isinstance(value, list):
             value.append(object)
             return value
        else:
            raise DebuggerCommandError("Invalid format")
    
    def sort(self, value:list, reverse=False):
        """
        Sort the value

        Args:
            value (list): List to sort.

        """
        if isinstance(value, list):
             sorted(value, reverse=reverse)
             return value
        else:
            raise DebuggerCommandError("Invalid format")

    def extend(self, value:list, iterable):
        """
        Extend the tail of the value

        Args:
            value (list): Value extend to.
            iterable (): iterable to extend to value

        """
        if isinstance(value, list):
             value.extend(iterable)
             return value
        else:
            raise DebuggerCommandError("Invalid format")


import pytest

from unittest.mock import Mock

from dAngr.cli.grammar.control_flow import ForLoop, IfThenElse
from dAngr.cli.grammar.parser import lex_input, parse_input
from dAngr.cli.grammar.expressions import BashCommand, Comparison, Dictionary, Expression, Listing, Literal, DangrCommand, Operator, Property, PythonCommand, Range, VariableRef
from dAngr.cli.grammar.script import Body, Script
from dAngr.cli.grammar.statements import Assignment
from dAngr.exceptions import ParseError


class TestLexer:

    def setup_method(self):
        pass
    def teardown_method(self):
        pass

    lex_neg_tests = [
    ]
    lex_tests = [
        "",
        "%a"
        'cmd_0_1_',
        '_cmd_0_2',
        'cmd3 0 1 2',
        "!print('x')",
        '$ls',
        '&cmd_0_1_',
        """
if a != "Hello from Bash":
    !print("Printing from Python: ", a)
    x = 0
    while x < max:
        !print("while: ",x)
        x = x + 1
    !print("Printing from Python: ", a)
    $echo "Hello from Bash" 
""",
"""
for i in range(10):
    !print(i)
""",
"""
for i,n in {"1":2, "3":4}:
    !print(i,n)
""",
"""
for i,n in [5,6,7,8,9]:
    !print(i,n)
""",
"""
if a != "Hello from Bash":
    !print("Printing from Python: ", a)
    x = 0
""",
"""[1,2,3,4,5]asd""",
'!("{:08x} {:08x} {:08x}".format(int(&sym.password0,16), int(&sym.password1,16), int(&sym.password2,16)))'
    ]

    lex_compound_tests = [
        
    ]

    
    def test_lex_false(self):
        for a in self.lex_neg_tests:
            #must raise exception
            with pytest.raises(ParseError) as exc_info:
                r = lex_input(a)
            assert exc_info is not None, f"Expected exception when lexing {a}, got {r}"

    
    def test_lex_true(self):
        for a in self.lex_tests:
            #must work
            r = lex_input(a)
            assert r is not None

    
    def test_lex_compound(self):
        for a in self.lex_compound_tests:
            #must work
            r = lex_input(a)
            print(r)
            assert r is not None

    parse_neg_tests = [
        "%a"
    ]

    parse_tests = {
        "" : [],
        'run': [DangrCommand('run',None)],
        'run 0 1 2': [DangrCommand('run', None, Literal(0), Literal(1), Literal(2))],
        "!print(&vars.x)" : [PythonCommand(Literal("print("), VariableRef(Literal('x')), Literal(")"))],
        "!print('x')" : [PythonCommand(Literal("print('x')"))],
        '$ls' : [BashCommand(Literal('ls'))],
        '&run' : [DangrCommand('run', None)],
        'run "4" "3" "4"': [DangrCommand('run', None, Literal('4'), Literal('3'), Literal('4'))],
        'run "5" !(print(1)) "4"': [DangrCommand('run', None, Literal('5'), PythonCommand(Literal("print(1)")), Literal('4'))],
        'run "6" $(ls -la) "4"': [DangrCommand('run', None, Literal('6'), BashCommand(Literal("ls -la")), Literal('4'))],
        'run "7" !(print($(ls -la))) "4"': [DangrCommand('run', None, Literal('7'), PythonCommand(Literal("print("),BashCommand(Literal("ls -la")), Literal(")")), Literal('4'))],
"""
for i in range(10):
    !print(&vars.i)
""": [ForLoop(Range(Literal(10)), Body([PythonCommand('print(',VariableRef(Literal('i')),')')]),VariableRef(Literal('i')))],
"""
for i,n in {"1":2, "3":4}:
    !print(&vars.i,&vars.n)
""": [ForLoop(Dictionary({"1": Literal(2), "3": Literal(4)}), Body([PythonCommand('print(',VariableRef(Literal('i')),',',VariableRef(Literal('n')),')')]),VariableRef(Literal('n')), VariableRef(Literal('i')))],
    '{"1":2, "3":4}': [Dictionary({"1": Literal(2), "3": Literal(4)})],
"""
for i,n in [5,6,7,8,9]:
    !print(&vars.i,&vars.n)
""": [ForLoop(Listing([Literal(5), Literal(6), Literal(7), Literal(8), Literal(9)]), Body([PythonCommand('print(',VariableRef(Literal('i')),',',VariableRef(Literal("n")),')')]),VariableRef(Literal('n')), VariableRef(Literal('i')))],
"""
if a != "Hello from Bash":
    !print("Printing from Python: ", &vars.a)
    x = 0
""":[IfThenElse(Comparison(VariableRef(Literal('a')), Operator.NEQ, Literal('Hello from Bash')), Body([PythonCommand('print("Printing from Python: ", ', VariableRef(Literal("a")),')'), Assignment(VariableRef(Literal("x")),Literal(0))]))],
"""[1,2,3,4,5]""":[Listing([Literal(1), Literal(2), Literal(3), Literal(4), Literal(5)])],
'!print("Printing from Python: ", &vars.a)': [PythonCommand('print("Printing from Python: ", ', VariableRef(Literal("a")),')')],
"static i = 0": [Assignment(VariableRef(Literal('i'),True), Literal(0))],
'!("{:08x} {:08x} {:08x}".format(int(&vars.password0,16), int(&vars.password1,16), int(&vars.password2,16)))':[PythonCommand('"{:08x} {:08x} {:08x}"','.format(int(', VariableRef(Literal('password0')),',16), int(', VariableRef(Literal('password1')),',16), int(', VariableRef(Literal('password2')),',16))')],
'a2 = (get_symbol &vars.password2 "int")': [Assignment(VariableRef(Literal('a2')), DangrCommand('get_symbol', None, VariableRef(Literal('password2')), Literal('int')))],
'a2 = &(get_symbol password2 "int")': [Assignment(VariableRef(Literal('a2')), DangrCommand('get_symbol', None, VariableRef(Literal('password2')), Literal('int')))],
'set_entry_state add_options=[options.LAZY_SOLVES]': [DangrCommand('set_entry_state', None, add_options= Listing([Property(VariableRef(Literal('options')), 'LAZY_SOLVES')]))],
    }

    def test_parse_false(self):
        for a in self.parse_neg_tests:
            #must raise exception
            with pytest.raises(ParseError) as exc_info:
                r = parse_input(a, None)
            assert exc_info is not None, f"Expected exception when parsing {a}, got {r}"
            

    def test_parse_true(self):
        for a in reversed(self.parse_tests):
            r = parse_input(a, None)
            assert Script(self.parse_tests[a],[]) == r, f"Expected {Script(self.parse_tests[a],[])}, got {r}"

    def test_parse_def(self):
        input = """
def my_function(max, min):
    !print(max)
    !print("Hello from Python")
    r = 5
    a = $(echo "Hello from Bash")
    if a != "Hello from Bash":
        !print("Printing from Python: ", a)
        !print("ok")
    else:
        !print("Printing from Python-else: ", a)
    x = 0
    while x < max:
        !print("while: ",x)
        x = x + 1
    !print("Printing from Python: ", a)
    $echo "Hello from Bash"
    
max = 5
x = 0
my_function max x
!print(x)
!print(6*5)
"""
        r = parse_input(input, None)
        assert r is not None


                